# Copyright (c) 2023 John Heintz, Gist Labs https://gistlabs.com
# License Apache v2 http://www.apache.org/licenses/

import numpy.random
import pytest
from stablerandom import stablerandom
from stablerandom.stablerandom import _randomLocalStack

def random():
    return numpy.random.triangular(1,5,10)

@stablerandom
def stable():
    return numpy.random.triangular(1, 5, 10)

@stablerandom
def nested():
    return stable(), stable(), random()

def test_random_is_random():
    assert random() != random()

def test_stable_is_stable():
    assert stable() == stable()

def test_stable_normal():
    @stablerandom
    def stable_normal():
        return numpy.random.normal(10, 5)

    assert stable_normal() == stable_normal()

def test_stable_pareto():
    @stablerandom
    def stable_pareto():
        return numpy.random.pareto(10)

    assert stable_pareto() == stable_pareto()

def test_stable_uniform():
    @stablerandom
    def stable_uniform():
        return numpy.random.uniform(0, 1)

    assert stable_uniform() == stable_uniform()


# Copyright (c) 2023 John Heintz, Gist Labs https://gistlabs.com
# License Apache v2 http://www.apache.org/licenses/
import numpy.random
from numpy.random import Generator, PCG64
import threading


class RandomLocal(threading.local):
    '''A threadlocal to hold a stack of stable randome generators'''
    stack = []

    def push(self):
        '''Push a new Generator PCG64 SeedSequence to the stack'''
        self.stack.append(Generator(PCG64(123456789)))

    def pop(self):
        '''Removes the top element from the stack'''
        self.stack.pop(0)

    def top(self) -> numpy.random.Generator:
        '''Lookup the top random generator, or return None'''
        if len(self.stack) > 0:
            return self.stack[-1]
        else:
            return None

_randomLocalStack = RandomLocal()

def stablerandom(func):
    """@stablerandom can be used stable random behavior for a function and call stack

    Usage:
        @stable_random
        def my_method():
            ...
    """
    def wrapper(*args, **kwargs):
        _randomLocalStack.push()

        try:
            return func(*args, **kwargs)
        finally:
            # clear thread local
            _randomLocalStack.pop()

    return wrapper


#random functions to wrap for stable,
# from https://numpy.org/doc/stable/reference/random/legacy.html#functions-in-numpy-random
_random_functions = ('normal', 'pareto', 'triangular', 'uniform')
def _wrap_numpy_random(funcName):
    _orig = getattr(numpy.random, funcName)

    def _stable(*args, **kwargs):
        stable = _randomLocalStack.top()
        if stable:
            func = getattr(stable, funcName)
            return func(*args, **kwargs)
        else:
            return _orig(*args, **kwargs)

    setattr(numpy.random, funcName, _stable)
[_wrap_numpy_random(f) for f in _random_functions]